/*
 * Handles the incoming signal as specified in the config file.
 *
 * (c)1999-2000 Frank van Viegen
 * This code is subject to the GPL. Read COPYING for details.
 */



#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <sched.h>
#include <time.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>

#include "irmanager.h"
#include "irman.h"
#include "codehandle.h"
#include "configparse.h"
#include "options.h"
#include "ioselect.h"

static void writeFifo(char *fifoname,char *t,int n);
static char *findSignal(ircode *code);
static struct keystr *findKey(struct mapstr *m, char *s);
static int readGarbage(int fd, int id);
static int setGarbageEater(int id);
int unblockLastCode(int id);
int setDefaultMap(int id);
int execActionList(struct keyactionstr *action, ircode *code);


struct mapstr *currentMap = NULL;
ircode lastCode;


/* This could write to multiple fifo's with the same internal name. */
static void writeFifo(char *fifoname,char *t,int n)
{
	struct fifostr *f;
	f = firstFifo;
	while(f!=NULL) {
		if (!strcasecmp(f->name,fifoname)) {
			unregFileSelect(f->file,IOSEL_CODEHANDLE);
			unregTimeSelect(f->file+1024);
			regTimeSelect(250, setGarbageEater, f->file+1024);
			write(f->file,t,n);
		}
		f = f->next;
	}
}

void initCodeHandle()
{
	struct fifostr *f;
	struct mapstr *m;
	struct stat filestat;
	int c;

	currentMap = defaultMap;
	if (defaultMap->expire || defaultMap->terminate) fatalerror("default map may not expire or terminate");
	
	unblockLastCode(0);

	f = firstFifo;
	umask(0);
	while(f!=NULL) {
		if (f->type==2 || (!optusefifo && f->type==0)) {
			if (!lstat(f->filename, &filestat)) {
				if (S_ISLNK(filestat.st_mode) || S_ISFIFO(filestat.st_mode)) unlink(f->filename);
				else fatalerror("fifo already exists as a regular file - i won't delete it");
			}
			if ((f->file = getpt())<0) fatalerror("unable to allocate pseudo terminal");
			grantpt(f->file);
			unlockpt(f->file);
			/* We'll open the slave too, so we won't get any trouble at the master-side when
			 * the slave is closed. */
			f->slavefile = open((char *)ptsname(f->file), O_RDWR);
			if (symlink((char *)ptsname(f->file),f->filename)) fatalerror("unable to create symlink to pseudo terminal");
		}
		else {
			if (!lstat(f->filename, &filestat) && !S_ISFIFO(filestat.st_mode)) {
				if (S_ISLNK(filestat.st_mode)) unlink(f->filename);
			}
			else {
				mkfifo(f->filename,6*8+6*64);
				if (lstat(f->filename, &filestat)) fatalerror("unable to create fifo");
			}
			if ((f->file = open(f->filename,O_RDWR|O_NONBLOCK))<0) fatalerror("unable to open fifo");
		}
		
		c=regFileSelect(SELECT_READ, f->file, readGarbage, IOSEL_CODEHANDLE);
		if (c) fatalerror("Fifo select error (%s)",strerror(c));
		f = f->next;
	}

	signal(SIGCHLD,SIG_IGN); /* Prevent zombies */
	
}

static int setGarbageEater(int id)
{
	regFileSelect(SELECT_READ, id-1024, readGarbage, IOSEL_CODEHANDLE);
	return(0);
}

static int readGarbage(int fd, int id)
{
	char c;
	read(fd,&c,1);
	return(0);
}

void finishCodeHandle()
{
	struct fifostr *f;
	unregFileSelect(-1,IOSEL_CODEHANDLE);
	f = firstFifo;
	while(f) {
		if (!close(f->file) && !close(f->slavefile) && !optusefifo) unlink(f->filename);
		f = f->next;
	}
}


static char *findSignal(ircode *code)
{
	struct codestr *s;
	s = firstCode;
	while (s!=NULL) {
		if (!codecmp(code,&s->code)) return (s->name);
		s = s->next;
	}
	return(NULL);
}


static struct keystr *findKey(struct mapstr *m, char *s)
{
	struct keystr *k;
	struct keystr *def;
	def = NULL;
	k = m->key;
	while (k!=NULL) {
		if (s && k->codename && !strcasecmp(k->codename,s)) return(k);
		if (k->codename==NULL) def = k;
		k = k->next;
	}
	return(def);
}

int setDefaultMap(int id)
{
	currentMap = defaultMap;
	return(0);
}

int unblockLastCode(int id)
{
	codecpy(&lastCode,"\0\0\0\0\0\0");
	return(0);
}

int handleCode(ircode *code)
{
	char *s, *t;
	struct keystr *key;
	int exit=0;
	

	/* Filter out repeated keys */
	if (optrepeatdelay) {
		if (!codecmp(&lastCode,code)) return(0);
		unregTimeSelect(IOSEL_KEYUNBLOCK);
		codecpy(&lastCode,code);
		if (optrepeatdelay>0) regTimeSelect(optrepeatdelay, unblockLastCode, IOSEL_KEYUNBLOCK);
	}
	
	/* Add a name to the signal */
	s = findSignal(code);

	/* Find an appropriate key with the name */
	key = findKey(globalMap,s);
	if (key) {
		if (!s) {
			t = code2text(code);
			loginfo("    Signal global.? (%s)\n", t)
			free(t);
		}
		else loginfo("    Signal global.%s\n",s);
	}
	else {
		key = findKey(currentMap,s);
		if (!key) {
			if (!s) {
				t = code2text(code);
				loginfo("    Signal ?.? (%s)\n", t)
				free(t);
			}
			else loginfo("    Signal ?.%s\n", s);
			return(0);
		}
		if (!s) {
			t = code2text(code);
			loginfo("    Signal %s.? (%s)\n", currentMap->name, t)
			free(t);
		}
		else loginfo("    Signal %s.%s\n", currentMap->name, s);
	}
	
	
	/* Mappings will only terminate on valid keys. */
	if (currentMap->terminate) currentMap=defaultMap;
	
	/* The current mapping does not need to expire any time soon if he
	 * was planning so, since a valid key was pressed.
	 */
	unregTimeSelect(IOSEL_MAPEXPIRE);
	
	/* Do not block this key if it is in direct mode. */
	if (key->direct) unblockLastCode(0);

	/* Execute the actions the key subscribes */
	execActionList(key->action,code);

	if (currentMap->expire) regTimeSelect(currentMap->expire*1000, setDefaultMap, IOSEL_MAPEXPIRE);

	return(exit);
}

/* Executes all commands in the linked list. Returns !0 if the exit command
 * has been given
 */
int execActionList(struct keyactionstr *action, ircode *code)
{
	struct mapstr *m;
	struct codestr *c;
	char *argv[4];
	int doexit=0;
	
	while (action!=NULL) {
		switch(action->cmdtype) {
		case CMD_SENDFIFO:
			if (action->op2) {
				loginfo("      Signal %s -> Fifo %s\n",action->op2,action->op1);
				c = firstCode;
				while (c) {
					if (!strcasecmp(c->name,action->op2)) {
						writeFifo(action->op1,c->code,6);
						break;
					}
					c = c->next;
				}
				warning("signal '%s' does not exist",action->op2);
			}
			else {
				if (!code) fatalerror("cannot send unspecified signal through Fifo from init");
				loginfo("      Fifo %s\n",action->op1);
				writeFifo(action->op1,(char *)code,6);
			}
			break;
		case CMD_WRITEFIFO:
			loginfo("      String '%s' -> Fifo %s\n",action->op2,action->op1);
			writeFifo(action->op1,action->op2,strlen(action->op2));
			break;
		case CMD_ACTIVATE:
			loginfo("      Activate %s\n",action->op1);
			m = firstMap;
			while (m!=NULL) {
				if (!strcasecmp(m->name,action->op1)) {
					currentMap=m;
					break;
				}
				m = m->next;
			}
			if (m==NULL) warning("map '%s' does not exist",action->op1);
			break;
		case CMD_EXIT:
			loginfo("      Exit\n");
			doexit=-1;
			break;
		case CMD_COMMAND:
			loginfo("      Execute %s\n",action->op1);
			argv[0] = "sh";
			argv[1] = "-c";
			argv[2] = action->op1;
			argv[3] = 0;
			if (!fork()) {
				if (logfd) stdout = stderr = logfd;
				else stdout = stderr = NULL;
				#ifndef DEBUGEXEC
				execv("/bin/sh",argv);
				#endif
			}
			break;
		}
		action = action->next;
	}
	return doexit;
}
