/*
 * Lexical analyzer for the config-file.
 *
 * (c)1999-2000 Frank van Viegen
 * This code is subject to the GPL. Read COPYING for details.
 */

%option yylineno
%x incl
%x comment

%{
	#include <stdio.h>
	#include <string.h>
	#include <stdlib.h>

	#include "irmanager.h"
	#include "irman.h"
	#include "config-gram.h"
	#include "configread.h"

	/* Including capability was copied from the Flex man page. */
	#define MAX_INCLUDE_DEPTH 10
	YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
	char *include_stack_name[MAX_INCLUDE_DEPTH+1] = {"top level"};
	int include_stack_lineno[MAX_INCLUDE_DEPTH+1] = {1};
	int include_stack_ptr = 0;
	int prevlineno = 1;
	char *configpath = NULL;
%}

%%

\/\/[^\n]*\n		{} 


"/*"			BEGIN(comment);
<comment>"*/"		BEGIN(INITIAL);
<comment><<EOF>>	{
			fatalerror("unexpected EOF in multiline comment %s",pos2text());
			}
<comment>\/\*		{
			fatalerror("nested comments are not permitted at %s",pos2text());
			}
<comment>[^\*\/]*	/* eat it */
<comment>\*[^\/]	/* this too */
<comment>\/[^\*]	/* and this */

"#include"		BEGIN(incl);

<incl>[ \t]*		/* eat the whitespace */

<incl>\"[^\"]*\"	{
			char *s, *t;
			int sl;

			yyin = 0;

			if (include_stack_ptr >= MAX_INCLUDE_DEPTH)
				fatalerror("too many layers of include used in config-file: %s",pos2text());

			sl = strlen(yytext);
			s = malloc(sl-2+1);
			memcpy(s, yytext+1, sl-2);
			*(s+sl-2) = '\0';

			if (s[0]!='/') {
				if (configpath) {
					t = malloc(strlen(s)+strlen(configpath)+1);
					strcpy(t,configpath);
				}
				else {
					t = malloc(strlen(s)+1);
				}
				strcat(t,s);
				yyin = fopen(t, "r");
				free(t);
			}
			if (!yyin) yyin = fopen(s, "r");
			if (!yyin) fatalerror("unable to open '%s' at %s", s, pos2text());


			include_stack[include_stack_ptr] = YY_CURRENT_BUFFER;
			include_stack_lineno[include_stack_ptr] += (yylineno-prevlineno);
			prevlineno = yylineno;
			include_stack_ptr++;
			include_stack_lineno[include_stack_ptr] = 1;

			setConfigName(s);
			free(s);

			yy_switch_to_buffer( yy_create_buffer(yyin,YY_BUF_SIZE) );

			BEGIN(INITIAL);
			}

<<EOF>>			{
			if (--include_stack_ptr<0) yyterminate();
			else {
				prevlineno = yylineno;
				yy_delete_buffer(YY_CURRENT_BUFFER);
				yy_switch_to_buffer(include_stack[include_stack_ptr]);
			}
			}

[0-9]+			{ yylval.valint = atoi(yytext); return NUMINT; }

"0x"[0-9a-f]+ 		{ yylval.valstr = (char *)text2code(yytext); return CODE; }

defaultmap		{ return DEFAULTMAP; }
map			{ return MAP; }
key			{ return KEY; }
command			{ return COMMAND; }
expire			{ return EXPIRE; }
terminate		{ return TERMINATE; }
sendfifo		{ return SENDFIFO; }
signal			{ return SIGNAL; }
fifo			{
			warning("deprecated keyword FIFO used at %s",pos2text());
			return FIFO;
			}
pty			{ return PTY; }
namedpipe		{ return NAMEDPIPE; }
true			{ return TRUE; }
false			{ return FALSE; }
activate		{ return ACTIVATE; }
direct			{ return DIRECT; }
serial			{ return SERIAL; }
exit			{ return EXIT; }
writefifo		{ return WRITEFIFO; }
init			{ return INIT; }

\=			{ return ('='); }
\*			{ return ('*'); }
\;			{ return (';'); }
\{			{ return ('{'); }
\}			{ return ('}'); }

\"[^\"]*\"		{
			int i;
			i = strlen(yytext);
			yylval.valstr = malloc(i-1);
			strncpy(yylval.valstr,yytext+1,i-2);
			yylval.valstr[i-2] = (char)0;
			return STRING;
			}

[ \t\r\n]+		{}

[a-z][a-z0-9_\-]*	{ yylval.valstr = strdup(yytext); return SYMBOL; }

<*>.			{ fatalerror("unrecognized symbol '%s' in %s",yytext,pos2text()); }

%%


char *pos2text()
{
	int i, size=0;
	char *s, *t;
	for (i=0;i<=include_stack_ptr;i++) {
		size += 2+1+2+5 + strlen(include_stack_name[i]); /* assume max 5 pos for lineno. DIRTY! */
	}
	size--;
	t = s = malloc(size);
	for (i=0;i<include_stack_ptr;i++) {
		t += sprintf(t,"(%s:%d)->",include_stack_name[i],include_stack_lineno[i]);
	}
	sprintf(t,"(%s:%d)",include_stack_name[i],include_stack_lineno[i]+yylineno-prevlineno);
	return(s);
}

void setConfigName(char *n)
{
	int c;
	for (c=strlen(n)-1;c>=0;c--) {
		if (n[c]=='/') {
			include_stack_name[include_stack_ptr] = strdup(n+c+1);
			if (!include_stack_ptr) {
				configpath = malloc(c+2);
				memcpy(configpath,n,c+1);
				configpath[c+1] = '\0';
			}
			return;
		}
	}
	include_stack_name[include_stack_ptr] = strdup(n);
}

int yywrap()
{
    return(-1);
}

