/*
 * This is sort of a convenient wrapper for select. I've written it for
 * IRManager, but I think it might prove useful for other things as well.
 *
 * (c)1999-2000 Frank van Viegen
 * This code is subject to the GPL. Read COPYING for details.
 */


/*
 * Restriction: I don't think it's safe yet for a function to remove its 
 * caller-structure. Fix it!
 */

#include <stdlib.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include "ioselect.h"

#include "irmanager.h"

#define OPERATIONS (SELECT_READ & SELECT_WRITE & SELECT_EXCEPT)

struct regFileChain {
	int fd;
	int id;
	int operations;
	int (*handler)(int,int);
	struct regFileChain *next;
};

struct regTimeChain {
	int deltatime;
	int id;
	int (*handler)(int);
	struct regTimeChain *next;
};

struct regSigChain {
	int signum;
	int id;
	int (*handler)(int,int);
	struct regSigChain *next;
};


static void defaultSignalHandler(int signum);


static struct regFileChain *fchain = NULL;
static struct regTimeChain *tchain = NULL;
static struct regSigChain *schain = NULL;
static unsigned int pendingSignals = 0;

/* the function registered here (only one) is called right after a signal
 * handler for the proces has run. i guess its main use is to let the mainloop
 * exit (by returning !0).
 * use NULL to unset the signal handler.
 */

void initIoSelect()
{
}

void finishIoSelect()
{
	int i;
	unregFileSelect(-1,-1);
	unregTimeSelect(-1);
	unregSignalSelect(-1,-1);
}

/* This can be used to handle signals synchronously.
 * In contrast to the other Selects, only one handler can be specified
 * for each signal.
 */
int regSignalSelect(int signum, int (*handler)(int,int), int id)
{
	struct sigaction t;
	struct regSigChain *n;
	if (signum>=32 || signum<0) return EINVAL;
	n = malloc(sizeof(struct regSigChain));
	if (!n) return ENOMEM;
	n->next = schain;
	n->id = id;
	n->signum = signum;
	n->handler = handler;
	sigaction(signum, NULL, &t);
	t.sa_handler = defaultSignalHandler;
	sigaction(signum, &t, NULL);
	schain = n;
	return 0;
}

static void defaultSignalHandler(int signum)
{
	pendingSignals |= (1<<(signum-1));
}

/* Use <0 for wildcards */
/* Returns the number of signals unselected */
int unregSignalSelect(int signum, int id)
{
	int count=0;
	struct regSigChain *n, *parent = NULL;
	struct sigaction t;
	for(n = schain; n; n = n->next) {
		if ((n->id==id || id<0) && (n->signum==signum || signum<0)) {
			count++;
			if (!parent) {
				schain = n->next;
				free(n);
				return(unregSignalSelect(signum,id)+count);
			}
			parent->next = n->next;
			free(n);
			n = parent;
		}
		else parent = n;
	}
	return count;
}


int regFileSelect(int operations, int fd, int (*handler)(int,int), int id)
{
	struct regFileChain *n;
//	if ((operations & (~OPERATIONS)) || id<0) return EINVAL;
	if (!operations) return 0;
	n = malloc(sizeof(struct regFileChain));
	if (!n) return ENOMEM;
	n->fd = fd;
	n->id = id;
	n->operations = operations;
	n->handler = handler;
	n->next = fchain;
	fchain = n;
	return(0);
}


/* fd's and id's < 0 will be considered a wildcard */
/* returns the number of entries removed or an error code < 0 */
int unregFileSelect(int fd, int id)
{
	int count=0;
	struct regFileChain *n, *parent = NULL;
	for(n = fchain; n; n = n->next) {
		if ((n->id==id || id<0) && (n->fd==fd || fd<0)) {
			count++;
			if (!parent) {
				fchain = n->next;
				free(n);
				return(unregFileSelect(fd,id)+count);
			}
			parent->next = n->next;
			free(n);
			n = parent;
		}
		else parent = n;
	}
	return count;
}


int regTimeSelect(int timemil, int (*handler)(int), int id)
{
	struct regTimeChain *n, *t, *parent=NULL;
	n = malloc(sizeof(struct regTimeChain));
	if (!n) return ENOMEM;
	n->id = id;
	n->handler = handler;
	if (tchain) {
		for(t = tchain; t; t = t->next) {
			if (t->deltatime>=timemil) {
				n->deltatime = timemil;
				n->next = t;
				t->deltatime -= timemil;
				if (parent) parent->next = n;
				else tchain = n;
				break;
			}
			timemil -= t->deltatime;
			parent = t;
		}
		if (!t) {
			n->next = NULL;
			n->deltatime = timemil;
			parent->next = n;
		}
	}
	else {
		n->next = NULL;
		n->deltatime = timemil;
		tchain = n;
	}
	return 0;
}


/* Remove timeSelect's with id. Use id<0 to remove all. */
int unregTimeSelect(int id)
{
	int count=0;
	struct regTimeChain *n, *parent = NULL;
	for(n = tchain; n; n = n->next) {
		if (n->id==id || id<0) {
			count++;
			if (!parent) {
				tchain = n->next;
				free(n);
				return(unregTimeSelect(id)+count);
			}
			parent->next = n->next;
			free(n);
			n = parent;
			count++;
		}
		else parent = n;
	}
	return count;
}


int ioSelect()
{
	fd_set readSet, writeSet, exceptSet;
	fd_set *readSetPtr=NULL-1, *writeSetPtr=NULL-1, *exceptSetPtr=NULL-1;
	struct timeval tv;
	struct regFileChain *nf;
	struct regTimeChain *nt;
	struct regSigChain *ns;
	struct sigaction sigstr;
	int mp, c, i, exit=0;
	int sigcheck=0;

	while(1) {
		mp=-1;
		if (readSetPtr) {
			FD_ZERO(&readSet);
			readSetPtr = NULL;
		}
		if (writeSetPtr) {
			FD_ZERO(&writeSet);
			writeSetPtr = NULL;
		}
		if (exceptSetPtr) {
			FD_ZERO(&exceptSet);
			exceptSetPtr = NULL;
		}
		for (nf=fchain; nf; nf=nf->next) {
			if (nf->operations & SELECT_READ) {
				FD_SET(nf->fd,&readSet); 
				readSetPtr = &readSet;
			}
			if (nf->operations & SELECT_WRITE) {
				FD_SET(nf->fd,&writeSet);
				writeSetPtr = &writeSet;
			} 
			if (nf->operations & SELECT_EXCEPT) {
				FD_SET(nf->fd,&exceptSet);
				exceptSetPtr = &exceptSet;
			}
			if (nf->fd>mp) mp = nf->fd;
		}
		
		if (tchain) {
			tv.tv_sec = tchain->deltatime / 1000;
			tv.tv_usec = (tchain->deltatime % 1000)*1000;
			c = select(mp+1, readSetPtr, writeSetPtr, exceptSetPtr, &tv);
		}
		else {
			if (!readSetPtr && !writeSetPtr && !exceptSetPtr && !schain) return 0;
			c = select(mp+1, readSetPtr, writeSetPtr, exceptSetPtr, NULL);
		}


		/* Check for errors */
		if (c<0) if (errno!=EINTR) return(errno);
		
		/* Have we hit the time-out? */
		if (c==0) tchain->deltatime = 0;
		/* If not, we might have to adjust our delta-table. This technique is
		 * pretty system specific, but a lot easier than its alternatives. 
		 */
		else if (tchain) tchain->deltatime = tv.tv_sec*1000 + tv.tv_usec/1000;
		

		/* Or are there any files read? */
		if (c>0) {
			for (nf=fchain; nf; nf=nf->next) {
				if ((nf->operations & SELECT_READ) && FD_ISSET(nf->fd,&readSet)) {
					exit |= nf->handler(nf->fd,nf->id);
				}
				if ((nf->operations & SELECT_WRITE) && FD_ISSET(nf->fd,&writeSet)) {
					exit |= nf->handler(nf->fd,nf->id);
				}
				if ((nf->operations & SELECT_EXCEPT) && FD_ISSET(nf->fd,&exceptSet)) {
					exit |= nf->handler(nf->fd,nf->id);
				}
			}
		}

		/* Process time-outs. */		
		while(tchain && !tchain->deltatime) {
			nt = tchain;
			tchain = tchain->next;
			exit |= nt->handler(nt->id);
			free(nt);
		}
		/* Process pending signals */
		while(pendingSignals) {
			if (pendingSignals&(1<<sigcheck)) {
				c=0;
				for(ns=schain; ns; ns = ns->next) {
					if (sigcheck+1==ns->signum) {
						c++;
						exit |= ns->handler(sigcheck+1,ns->id);
					}
				}
				if (!c) { /* we don't seem to handle this signal anymore */
					sigaction(sigcheck+1, NULL, &sigstr);
					sigstr.sa_handler = SIG_DFL;
					sigaction(sigcheck+1, &sigstr, NULL);
				}
				pendingSignals &= (~(1<<sigcheck));
			}
			else sigcheck = (sigcheck+1)&31;
		}
		
		if (exit) return(0);
		
	}
	return(0);
}
