/*
 * This module provides IRMan functionality, using the TTYIO interface to the serial port.
 *
 * It is strongly based (is it still? :) on code by Tom Wheeley. (thanks!)
 * The original code was LGPL, which makes this module the same.
 *
 * (c)2000 Frank van Viegen
 *
 */
 
/* xmms-irman v0.4.1 (c) 1998-99 Tom Wheeley <tomw@tsys.demon.co.uk> */
/* this code is placed under the LGPL, see www.gnu.org for info    */


#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/file.h>
#include <unistd.h>
#include <termios.h>

#include "irman.h"
#include "ttyio.h"
#include "irmanager.h"
#include "ioselect.h"
#include "options.h"

#define SSEC	* 1000000
#define MSEC	* 1000
#define USEC
#define IR_POWER_ON_LATENCY	(10 MSEC)
#define IR_HANDSHAKE_GAP	(500 USEC)
#define IR_GARBAGE_TIMEOUT	(50 MSEC)


static int hex2int(unsigned char hex);
int (*codeHandler)(ircode *) = NULL;
int irmanHandler(int fd, int id);
int keyboardHandler(int fd, int id);


static ircode newCode;
static int codeStatus;
static struct ttyInfo *tty;
static struct termios stdinOldTerm;


/* We don't need this anymore, since we'll be working event-driven. */
/*
 *int getIrcode(ircode *code)
 *{
 *	int i=0, t;
 *	while (i<IR_CODE_LEN) {
 *		t=tty_read_char(-1);
 *		if (t<0) {
 *			if (exitIrman) return -1;
 *			else fatalerror("unable to read from IRMan device");
 *		}
 *		((char *)code)[i] = (char)t;
 *		i++;
 *	}
 *	return 0;
 *}
 */


void initIrman(char *f, int (*handler)(ircode *))
{
	int i,c;
	codeHandler = handler;
	/* If needed, rehearse a couple of times :) */
	for(i=0;i<3;i++) {
		c=0;
		if ( c==0 && (optnolocking || !lockTty(f)) ) c++;
		if (c==1 && (tty=openTty(f))) c++;
		if (c==2) {
			flushIrman();
			ttyWriteChar(tty,'I');
			ttyUsleep(IR_HANDSHAKE_GAP);
			ttyWriteChar(tty,'R');
			ttyUsleep(IR_POWER_ON_LATENCY);
			if (ttyReadChar(tty,50 MSEC)=='O' && ttyReadChar(tty,50 MSEC)=='K') c++;
		}
		if (c==3) break;
		if (c>1) closeTty(tty);
		if (c>0 && !optnolocking) unlockTty(f);
		ttyUsleep(250 MSEC);
	}
	switch(c) {
	case 0: fatalerror("unable to lock tty '%s' - might another instance of IRManager be running?",f);
	case 1: fatalerror("unable to open tty '%s'",f);
	case 2: fatalerror("unable to initialize IRMan on tty '%s'",f);
	}
	codeStatus = 0;
	i=regFileSelect(SELECT_READ, tty->fd, irmanHandler, IOSEL_IRMAN);
	if (i) fatalerror("irman init select error (%s)",strerror(i));
}

void initKeyboard(int (*handler)(ircode *))
{
	int i;
	struct termios t;
	codeHandler = handler;
	tcgetattr(0,&t);
	memcpy(&stdinOldTerm,&t,sizeof(struct termios));
	t.c_lflag &= ~(ECHO | ICANON);
	tcsetattr(0,TCSANOW,&t);
	i=regFileSelect(SELECT_READ, 0, keyboardHandler, IOSEL_KEYBOARD);
	if (i) fatalerror("keyboard init select error (%s)",strerror(i));
}

int keyboardHandler(int fd, int id)
{
	char x;
	int i;
	read(fd,&x,1);
	for (i=0;i<sizeof(ircode);i++) {
	    *((char *)&newCode+i) = x;
	}
	return(codeHandler(&newCode));
}

int irmanHandler(int fd, int id)
{
	if(read(fd,((char *)&newCode)+codeStatus,1)>0) {
		codeStatus++;
		if (codeStatus>=sizeof(ircode)) {
			codeStatus=0;
			return(codeHandler(&newCode));
		}
	}
	return(0);
}

void finishIrman()
{
	char *tn;
	unregFileSelect(-1,IOSEL_IRMAN);
	tn = strdup(tty->name);
	closeTty(tty);
	unlockTty(tn);
	free(tn);
}

void finishKeyboard()
{
	tcsetattr(0,TCSANOW,&stdinOldTerm);
	unregFileSelect(-1,IOSEL_KEYBOARD);
}

void flushIrman()
{
	while (ttyReadChar(tty,IR_GARBAGE_TIMEOUT) >= 0);
}



/* output hex digits */
static char ir_hexdigit[16] =
{
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'
};


char *code2text(ircode *code)
{
	char *text, *j;
	int i;
	
	text = (char *)malloc(2*IR_CODE_LEN+3);

	j = text;
	*j++ = '0'; *j++ = 'x';
	for (i = 0; i < IR_CODE_LEN; i++)
	{
		*j++ = ir_hexdigit[(((char *)code)[i] >> 4) & 0x0f];
		*j++ = ir_hexdigit[((char *)code)[i] & 0x0f];
	}
	*j = '\0';

	return text;
}

static int hex2int(unsigned char hex)
{
	if (hex >= '0' && hex <= '9')
		return hex - '0';

	hex = tolower(hex);
	if (hex >= 'a' && hex <= 'f')
		return hex - 'a' + 10;

	/* error! */
	return -1;
}

ircode *text2code(char *tstart)
{
	char *cstart, *twalk, *cwalk;
	int fase=0,t;

	cstart = (char *)malloc(IR_CODE_LEN);
	if (!cstart) fatalerror("not enough memory");
	for (cwalk=cstart;cwalk<cstart+IR_CODE_LEN;cwalk++) *cwalk = (char)0;
	cwalk--;

	if (tstart[0]=='0' && tstart[1]=='x') tstart+=2;
	twalk = tstart+strlen(tstart)-1;

	while(cwalk>=cstart && twalk>=tstart)
	{
		t = hex2int(*twalk--);
		if (t<0) fatalerror("invalid signal '0x%s'",tstart);
		if (!fase) *cwalk = (char)t & 0x0f;
		else *(cwalk--) += (char)(t<<4) & 0xf0;
		fase = (fase+1) & 1;
	}
	return (ircode *)cstart;
}
