/*
 * This module provides reliable access to serial devices.
 *
 * It is strongly based on code by Tom Wheeley. (thanks!)
 * The original code was LGPL:
 *
 *   xmms-irman v0.4.1 (c) 1998-99 Tom Wheeley <tomw@tsys.demon.co.uk>
 *   this code is placed under the LGPL, see www.gnu.org for info
 *
 * The tty-locking code was borrowed from Heartbeat and written by
 * Guenther Thomsen. (also thanks!) This code is GPL. Now I'm not
 * sure about the license of the module, as it includes GPL as well
 * as LGPL code from others. I guess it's still GPL. Or am I
 * violating any licenses here?
 *
 * (c)2000 Frank van Viegen
 *
 */



#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <termios.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include "ttyio.h"
#include "irmanager.h"

/*
 * Note regarding terminal settings.
 *
 * These work on my system.  I am quite confident they will work on other
 * systems.  The termios setup code is originally from another program
 * designed to talk to a serial device (casio diary) written by someone who I
 * can't remember but I presume they knew what they were doing.
 *
 * More information on Unix serial port programming can be obtained from
 *   http://www.easysw.com/~mike/serial/index.html
 *
 * addendum.  I tried for quite a while to get this all to work on a Sun
 * Ultra 1, but it didn't :(  All the attempts are under ifdef SUNATTEMPT.
 */

/*
 * A future avenue might be to set the settings to 0 and work from there.
 * Also check for other names for the serial port as in IRIX -- maybe
 * Solaris does that (I tried ttya)
 */


struct ttyInfo *openTty ( char *filename )
{
	int parnum = 0;
	int baudrate = B9600;
	int flags;
	struct termios term;
	struct ttyInfo ti, *tip;

	/* open tty */
	if ( ( ti.fd = open ( filename, O_RDWR | O_NOCTTY | O_NDELAY ) ) < 0 )
		return NULL;

	/* is it a terminal? */
	if ( !isatty ( ti.fd ) )
		return NULL;

	/* get port attributes, store in oldterm */
	if ( tcgetattr ( ti.fd, &ti.oldterm ) < 0 )
		return NULL;

	/* get port flags, save in oldflags */
	if ( ( ti.oldflags = fcntl ( ti.fd, F_GETFL ) ) < 0 )
		return NULL;

	/* now we have read the old attributes for the port, we can restore them
	 * upon exit. if we had done this bfore, and exited beore reading in the
	 * old attributes, we would have overwritten the old settings with zeros.  
	 *
	 * this way, if we do exit before we get here, we simply rely on the OS closing
	 * the port for us, which is fine as we haven't changed anything yet.
	 */

	term = ti.oldterm;
	flags = ti.oldflags;

	/* remove old parity setting, size and stop setting */
	term.c_cflag &= ~PARENB;
	term.c_cflag &= ~PARODD;
	term.c_cflag &= ~CSTOPB;
	term.c_cflag &= ~CSIZE;

	/* set character size, stop bits and parity */
	term.c_cflag |= CS8;
	term.c_cflag |= parnum;

	/* enable receiver, and don't change ownership */
	term.c_cflag |= CREAD | CLOCAL;

	/* disable flow control */
#ifdef CNEW_RTSCTS
	term.c_cflag &= ~CNEW_RTSCTS;
#else
#ifdef CRTSCTS
term.c_cflag &= ~CRTSCTS;
#endif
#ifdef CRTSXOFF
	term.c_cflag &= ~CRTSXOFF;
#endif
#endif

	/* read characters immediately in non-canonical mode */
	/* Thanks to Bill Ryder, <bryder@sgi.com> */
	term.c_cc[ VMIN ] = 1;
	term.c_cc[ VTIME ] = 1;

	/* set the input and output baud rate */
	cfsetispeed ( &term, baudrate );
	cfsetospeed ( &term, baudrate );

	/* set non-canonical mode (we don't want any fancy terminal processing!) */
	term.c_lflag = 0;

	/* Ignore breaks and make terminal raw and dumb. */
	term.c_iflag = 0;
	term.c_iflag |= IGNBRK;
	term.c_oflag &= ~OPOST;

	/* set the input and output baud rate */
	cfsetispeed ( &term, baudrate );
	cfsetospeed ( &term, baudrate );

	/* now clean the serial line and activate the new settings */
	tcflush ( ti.fd, TCIOFLUSH );
	if ( tcsetattr ( ti.fd, TCSANOW, &term ) < 0 )
		return NULL;

	/* set non-blocking */
	if ( fcntl ( ti.fd, F_SETFL, ( flags |= O_NONBLOCK ) ) < 0 )
		return 0;

	/* Thanks to mumble, < mumble @ mumble.gr > for the suggestion.
	 * (sorry I forgot your name/address-- email me and I'll put it in!)
	 */

	ti.name = strdup( filename );
	tip = malloc( sizeof ( struct ttyInfo ) );
	memcpy( tip, &ti, sizeof ( struct ttyInfo ) );

	return tip;
}


int closeTty ( struct ttyInfo *tip )
{
	int retval = 0;
	/* restore old settings */
	if ( tcsetattr ( tip->fd, TCSADRAIN, &tip->oldterm ) < 0 ) retval = -1;

	if ( fcntl ( tip->fd, F_SETFL, tip->oldflags ) < 0 ) retval = -1;

	close ( tip->fd );

	free ( tip->name );
	free ( tip );

	return retval;
}


/* write a character.  nothing interesting happens here */

int ttyWriteChar ( struct ttyInfo *tip, unsigned char data )
{
	if ( write ( tip->fd, &data, 1 ) != 1 ) return -1;
	else {
#ifdef DEBUG_COMM
		printf ( "{%02x}", data );
		fflush ( stdout );
#endif 
		return 0;
	}
}


/* read a character, with a timeout.
 * timeout < 0	-  block indefinitely
 * timeout = 0  -  return immediately
 * timeout > 0  -  timeout after `timeout' microseconds
 *                 use the nice macros in irman.h to define sec, msec, usec
 */

int ttyReadChar ( struct ttyInfo *tip, long timeout )
{
	unsigned char rdchar;
	int ok;
	fd_set rdfds;
	struct timeval tv;

	do {
		FD_ZERO ( &rdfds );
		FD_SET ( tip->fd, &rdfds );
		/* block until something to read or timeout occurs.  select() is damn cool */
		if ( timeout < 0 ) {
			ok = select ( tip->fd + 1, &rdfds, NULL, NULL, NULL );
		} else {
			tv.tv_sec = timeout / 1000000;
			tv.tv_usec = ( timeout % 1000000 );
			ok = select ( tip->fd + 1, &rdfds, NULL, NULL, &tv );
		}

	} while ( ok == -1 );

	if ( ok > 0 ) {
		ok = read ( tip->fd, &rdchar, 1 );
		if ( ok == 0 ) {
			return EOF;
		}
#ifdef COMM_DEBUG
		printf ( "[%02x]", rdchar );
		fflush ( stdout );
#endif 
		return rdchar;
	} else
		if ( ok < 0 ) {
			return EOF - 1;
		} else {
			errno = ETIMEDOUT;
			return EOF - 1;
		}

	return 0;
}


/* just about the only function where we don't care ir_about errors! */

/* some systems have it, some systems don't.  This just makes life easier,
 * hence I have left this function visible (also for irfunc.c)
 */

void ttyUsleep ( unsigned long usec )
{
	struct timeval tv;
	tv.tv_sec = usec / 1000000;
	tv.tv_usec = ( usec % 1000000 );
	( void ) select ( 0, NULL, NULL, NULL, &tv );
}



/* lock a tty (using lock files, see linux `man 2 open` close to O_EXCL)
 * serial_device has to be _the complete path_, i.e. including '/dev/' to the
 * special file, which denotes the tty to lock -tho
 * return 0 on success, 
 * -1 if device is locked (lockfile exists and isn't stale),
 * -2 for temporarily failure, try again,
 * other negative value, if something unexpected happend (failure anyway)
 */
 
#if !defined(__FreeBSD__)
	#define  TTY_LOCK_D      "/var/lock"
#else
	#define  TTY_LOCK_D      "/var/spool/lock"
#endif

int lockTty (char *ttydev)
{
	char lf_name[ 256 ], tf_name[ 256 ], buf[ 12 ];
	int fd;
	pid_t pid, mypid;
	int rc;
	mode_t oldumask;
	struct stat sbuf;
	
	if (access(TTY_LOCK_D,W_OK|R_OK|X_OK)) {
		return -1;
	}
	
	mypid = getpid ();
	snprintf ( lf_name, 256, "%s/LCK..%s", TTY_LOCK_D,
	           ttydev + sizeof ( "/dev/" ) - 1 );
	snprintf ( tf_name, 256, "%s/tmp%d-%s", TTY_LOCK_D,
	           mypid,
	           ttydev + sizeof ( "/dev/" ) - 1 );
	if ( ( fd = open ( lf_name, O_RDONLY ) ) >= 0 ) {
		if ( read ( fd, buf, 12 ) < 1 ) {
			/* lockfile empty -> rm it and go on */
		} else {
			if ( sscanf ( buf, "%d", &pid ) < 1 ) {
				/* lockfile screwed up -> rm it and go on */
			} else {
				if ( !kill(pid, 0) || errno != ESRCH ) {
					/* tty is locked by existing (not
					 * necessarily running) process
					 * -> give up */
					close ( fd );
					return -1;
				} else {
					/* stale lockfile -> rm it and go on */
				}
			}
		}
		unlink ( lf_name );
	}
	
	oldumask = umask(0);
	fd = open ( tf_name, O_CREAT | O_WRONLY, 0664 );
	umask(oldumask);
	if (fd<0) {
		/* Hmmh, why did we fail? Anyway, nothing we can do about */
		return -3;
	}
	sprintf ( buf, "%10d\n", mypid );
	if ( write ( fd, buf, 12 ) < 12 ) {
		/* again, nothing we can do about */
		return -3;
	}
	close ( fd ); switch ( link ( tf_name, lf_name ) ) {
	case 0:
		if ( stat ( tf_name, &sbuf ) < 0 ) {
			/* something weird happened */
			rc = -3;
			break;
		}
		if ( sbuf.st_nlink < 2 ) {
			/* somehow, it didn't get through - NFS trouble? */
			rc = -2;
			break;
		}
		rc = 0;
		break;
	case EEXIST:
		rc = -1;
		break;
	default:
		rc = -3;
	}
	unlink ( tf_name ); return rc;
}

/* unlock a tty (remove its lockfile)
 * do we need to check, if its (still) ours? No, IMHO, if someone else
 * locked our line, it's his fault  -tho
 * returns 0 on success
 * <0 if some failure occured */
int unlockTty (char *ttydev)
{
	char lf_name[ 256 ];
	snprintf ( lf_name, 256, "%s/LCK..%s", TTY_LOCK_D,
	           ttydev + sizeof ( "/dev/" ) - 1 );
	return unlink ( lf_name );
}
